import 'package:collection/src/iterable_extensions.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ConverterGenerator {

  final Schema schema;
  ConverterGenerator(this.schema);

  Iterable<Model> get models => schema.models;

  String compile() =>
'''
class _Converter extends Converter {
  const _Converter();

  @override
  StateValue decodePut(DataRecord r) {
    final data = r.toJsonData();
    switch(r.type) {${models.map((m) =>'''
      case '${m.name}': return ${m.putDecoder};''').join()}
      default: throw Exception('unknown type \${r.type} during decodePut');
    }
  }

  @override
  StateKey decodeRemove(DataRecord r) {
    switch(r.type) {${models.map((m) =>'''
      case '${m.name}': return ${m.removeDecoder};''').join()}
      default: throw Exception('unknown type \${r.type} during decodeRemove');
    }
  }

  @override
  DataRecord encodePut(StateValue v) {${models.map((m) =>'''
    if(v is ${m.snapName}) {
      return ${m.putEncoder};
    }''').join(' else ')}
    else {
      throw Exception('unknown type \$v during encodePut');
    }
  }

  @override
  DataRecord encodeRemove(StateValue v) {${models.map((m) =>'''
    if(v is ${m.snapName}) {
      return ${m.removeEncoder};
    }''').join(' else ')}
    else {
      throw Exception('unknown type \$v during encodeRemove');
    }
  }
}
''';
}

extension ModelX on Model {

  String get removeDecoder => isSingleton
      ? '$keyName()'
      : '$keyName.parse(r.modelId)';

  String get removeEncoder => isSingleton
      ? "DataRecord('$name', '$name', v.updateId.hexString)"
      : "DataRecord('$name', v.id.toString(), v.updateId.hexString)";

  String get putDecoder => isSingleton
      ? '$snapName._($fromRecordEntries, ObjectId.fromHexString(r.updateId))'
      : '$snapName._($keyName.value(r.modelId), $fromRecordEntries, ObjectId.fromHexString(r.updateId))';

  String get putEncoder => isSingleton
      ? "DataRecord('$name', '$name', v.updateId.hexString, {$toRecordEntries})"
      : "DataRecord('$name', v.id.toString(), v.updateId.hexString, {$toRecordEntries})";

  String get fromRecordEntries => dataFields.map((f) => f.isNullable
    ? "((data['${f.name}'] != null) ? ${f.fieldDecoder} : null)"
    : "${f.fieldDecoder}"
  ).join(',');

  String get toRecordEntries => dataFields.map((f) => f.isNullable
    ? "if(v.${f.snapName} != null) '${f.name}': ${f.fieldEncoder}"
    : "'${f.name}': ${f.fieldEncoder}"
  ).join(',');
}

extension ConverterFieldX on ModelField {

  String get fieldDecoder => _recordDecoder();
  String _recordDecoder([String? v]) {
    v ??= "data['$name']";
    return model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonDecoder?.replaceAll('\$', v)
        ?? (isClass ? '$type.fromJson($v)' : null)
        ?? ((isModel && typeModel!.isInternal) ? '${typeModel!.keyName}.value($v)' : null)
        ?? (isModel ? v : null)
        ?? (isHasMany ? (throw Exception('recordDecoder on isMany field')) : null)
        ?? (isDateTime ? 'DateTime.fromMillisecondsSinceEpoch($v)' : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? fromFieldCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String get fieldEncoder {
    final v = 'v.$snapName';
    final e = model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonEncoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.encoder?.replaceAll('\$', v)
        ?? (isHasMany ? (throw Exception('recordEncoder on isMany field')) : null)
        ?? (jsonEncode?.replaceAll('\$', v))
        ?? (isClass ? '$v$nullable.toJson()' : null)
        ?? ((isModel && typeModel!.isInternal) ? '$v$nullable.hexString' : null)
        ?? (isModel ? v : null)
        ?? (isObjectId ? '$v$nullable.hexString' : null)
        ?? (isDateTime ? '$v$nullable.millisecondsSinceEpoch' : null)
        ?? (isPrimitive ? v : null)
        ?? (isInPart ? '$v$nullable.toJson()' : null)
        ?? (isEnum ? '$v$nullable.name' : null)
        ?? 'recordValue($v)';
    return isNullable && isInitialized ? '$e ?? $initializer' : e;
  }

  String fromFieldCollection(String value) {
    if(isNotParameterized) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ModelField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final fromRecord = field._recordDecoder('v');
    if(isIterable) return '($value ?? []).map((v) => $fromRecord).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $fromRecord).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $fromRecord).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $fromRecord)).cast<$parameter>()';
    throw '$this is not a collection';
  }
}