import 'package:collection/src/iterable_extensions.dart';
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelGenerator {

  final Model m;
  Schema get schema => m.schema;
  ModelGenerator(this.m);

  String compile() => [
    compileModel,
    compileKey,
    compileSnap,
    compileRef,
    if(m.isNotSingleton) compileQuery,
  ].join('\n');

  String get compileKey =>
m.isSingleton ?
'''
class ${m.keyName} extends StateKey<${m.snapName}> {
  @override
  final String id = '${m.name}';
  const ${m.keyName}();
}
''' :
m.isInternal ?
'''
class ${m.keyName} extends StateKey<${m.snapName}> {
  @override
  final ObjectId id;
  ${m.keyName}([ObjectId? id]) : id = id ?? ObjectId();
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ObjectId.fromHexString(s);
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''' :
'''
class ${m.keyName} extends StateKey<${m.snapName}> {
  @override
  final ${m.idType} id;
  const ${m.keyName}(this.id);
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ${(m.idType == 'String') ? 's' : '${m.idType}.parse(s)'};
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''';

  String get compileModel =>
'''
abstract class ${m.name}$baseInterfaces{
  
  $abstractGetters
  
}
''';

  String get compileSnap =>
'''
class ${m.snapName} extends $snapSuper$mixins$modelInterfaces {
  
  ${m.snapName}($snapParams) : this._($snapConstructor);
  ${m.snapName}.fromJson(Map<String, dynamic> data, {$fromJsonParams}) : this._($fromJsonConstructor);
  static List<${m.snapName}> fromJsonList(list, {$fromJsonListParams}) => (list as Iterable).map((data) => ${m.snapName}.fromJson($fromJsonListConstructor)).toList();
  ${m.snapName}._($snap_Constructor)$snap_Super
  
  ${m.isSingleton ? "final String id = '${m.name}';" : ''}
  ${m.dataFields.map((f) => f.snapFieldDeclaration).join()}
  
  @override
  int compareTo(${m.snapName} other) => $compareToTests;
  
  ${m.isNotSingleton ? '@override' : ''}
  ${m.snapName} copyWith({$copyWithParams}) => ${m.snapName}._($copyWithConstructor);
  
  $toJson
}
''';

  String get compileRef =>
      m.isSingleton ? singletonRef :
      m.isExternal ? externalRef : internalRef;

  String get externalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $fetch
  $update
}
''';

  String get internalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $update
}
''';

  String get singletonRef =>
'''
class $declaration {
  $constructor
  final String id = '${m.name}';
  $accessors
  $fetch
  $update
}
''';

  String get compileQuery =>
      '''
class ${m.queryName} {
  
  $queryFields
  
  const ${m.queryName}({$queryParams});
  ${m.queryName}._($queryParams);
  
  bool get isAll => $queryAll;
  bool get isNotAll => !isAll;
  
  bool test(${m.snapName} m) => $queryTest;

  @override
  bool operator ==(Object? other) => identical(this, other)
    || (other is ${m.queryName} && $equalsTests);
  
  @override
  int get hashCode => $objectHash;
}
''';

  String get queryFields => m.realFields.map((f) => [
    'final ${f.typeModel?.idType ?? f.type}? ${f.snapName};',
    if(f.isNullable) 'final bool? ${f.snapName}IsNull;'
  ]).expand((e) => e).join();

  String get queryParams => m.realFields.map((f) => [
    if(f.isNullable) 'this.${f.snapName}IsNull',
    'this.${f.snapName}',
  ]).expand((e) => e).join(',');

  String get queryAll => m.realFields.map((f) => [
    if(f.isNullable) '(${f.snapName}IsNull == null)',
    '(${f.snapName} == null)',
  ]).expand((e) => e).join('&&');

  String get queryTest => m.realFields.map((f) => [
    if(f.isNullable) '('
                      '(${f.snapName}IsNull == true) ? (m.${f.snapName} == null) :'
                      '(${f.snapName}IsNull == false) ?'
                        '(${f.snapName} == null) ? (m.${f.snapName} != null) : (${f.snapName} == m.${f.snapName}) :'
                      '((${f.snapName} == null) || (${f.snapName} == m.${f.snapName}))'
                     ')'
    else '(${f.snapName} == null || ${f.snapName} == m.${f.snapName})',
  ]).expand((e) => e).join('&&');

  String get equalsTests => m.realFields.map((f) => '${f.snapName} == other.${f.snapName}').join('&&');

  String get objectHash => m.fields.length < 20
      ? 'Object.hash($hashValues)' : 'Object.hashAll([$hashValues])';

  String get hashValues => [
    m.name,
    ...m.realFields.map((f) => f.snapName)
  ].join(',');

  String get snapSuper => m.isSingleton
    ? 'Singleton<${m.snapName}>'
    : 'Model<${m.snapName}, ${m.idType}>';

  String get snap_Constructor => [
    if(m.isNotSingleton) '${m.idType} id',
    ...m.dataFields.map((f) => 'this.${f.snapName}'),
    '[ObjectId? _updateId]'
  ].join(',');

  String get snap_Super => ' : super(${[
    if(m.isNotSingleton) 'id',
    '_updateId',
  ].join(',')});';

  String get snapParams => '{'
    '${m.isSingleton ? '' : m.isInternal ? 'ObjectId? id,' : 'required ${m.idType} id,'}'
    '${m.dataFields.map((f) => '${f.isRequired ? 'required ${f.jsonType}' : '${f.jsonType}${f.nullable}'} ${f.snapName}${f.isInitialized ? '=${f.initializer}' : ''}').join(',')}'
  '}';

  String get snapConstructor => [
    if(m.isExternal) 'id'
    else if(m.isNotSingleton) 'id ?? ObjectId()',
    ...m.dataFields.map((f) => f.snapName)
  ].join(',');

  String get declaration => m.isSingleton
      ? '${m.refName} extends SingletonRef<${m.snapName}, ${m.refName}>$mixins$modelInterfaces'
      : '${m.refName} extends ModelRef<${m.snapName}, ${m.idType}, ${m.refName}>$mixins$modelInterfaces';

  String get baseInterfaces => m.interfaces.isEmpty ? '' : ' implements ${m.interfaces.join(',')}';
  String get mixins => m.mixins.isEmpty ? '' : ' with ${m.mixins.join(',')}';
  String get modelInterfaces => ' implements ${m.name}, ${m.keyName}';

  String get constructor =>
m.isSingleton ?
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature) : super(_feature);
''' :
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature, ${m.idType} id) : super(_feature, id);
''';

  String get abstractGetters => [
    if(m.isNotSingleton)
      m.idField!.abstractGetter,
    ...m.dataFields.map((f) => f.abstractGetter)
  ].join();

  String get accessors => m.fields.where((f) => f.isNotId)
      .map((f) => f.accessors).expand((e) => e).join();

  String get fetch => schema.has<RestFeature>() ?
'''

  void expire() => _feature<_RestFeature>().${m.restAdapterVarName}.expire($expireArgs);

  Future<${m.refName}> fetch({bool force=false}) => _feature<_RestFeature>().${m.restAdapterVarName}.fetch($fetchArgs).then((_) => this);
''' : '';

  String get expireArgs => m.isNotSingleton ? '${m.queryName}(id: id)' : '';

  String get fetchArgs => [
    if(m.isNotSingleton) '${m.queryName}(id: id)',
    'force'
  ].join(',');

  String get update =>
'''
  void update(${m.updateParams}) {
    final s = ${m.snapName}._(${m.updateConstructor});
    _feature.put(s);
  }
''';

  String get compareToTests => [
    ...m.dataFields.map((f) => 'compare(${f.snapName}, other.${f.snapName})'),
    '0',
  ].join('??');

  String get copyWithParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
         '${f.jsonType}? ${f.snapName}',
         if(f.isNullable) 'bool ${f.snapName}AsNull=false'
       ]).expand((e) => e)
    ].join(',');

  String get copyWithConstructor => [
    if(m.isNotSingleton) 'id ?? this.id',
    ...m.dataFields.map((f) => f.isNullable
        ? '${f.snapName}AsNull ? null : (${f.snapName} ?? this.${f.snapName})'
        : '${f.snapName} ?? this.${f.snapName}')
  ].join(',');

  String get fromJsonParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.snapName}',
      if(f.isNullable) 'bool ${f.snapName}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonConstructor => [
    if(m.isExternal) "id ?? (data['id'] as ${m.idType})"
    else if(m.isInternal) "id ?? ((data['id'] != null) ? ${m.keyName}.value(data['id']) : ObjectId())",
    ...m.dataFields.map((f) => (f.isNullable || f.isInitialized)
        ? "${fromJsonOverrides(f)} ?? ((data['${f.remote}'] != null) ? ${f.jsonDecoder} : ${f.initializer})"
        : "${fromJsonOverrides(f)} ?? ${f.jsonDecoder}")
  ].join(',');

  String fromJsonOverrides(ModelField f) => [
    f.snapName,
    if(m.isNotSingleton) 'withQuery?.${f.snapName}'
  ].join('??');

  String get fromJsonListParams => [
    if(m.isNotSingleton) '${m.idType} Function(Map<String, dynamic> item)? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.snapName}',
      if(f.isNullable) 'bool ${f.snapName}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonListConstructor => [
    'data',
    if(m.isNotSingleton) 'id: id?.call(data)',
    ...m.dataFields.map((f) => [
      '${f.snapName}: ${f.snapName}',
      if(f.isNullable) '${f.snapName}AsNull: ${f.snapName}AsNull'
    ]).expand((e) => e),
    if(m.isNotSingleton) 'withQuery: withQuery',
  ].join(',');

  String get toJson =>
  '@override '
  'Map<String, dynamic> toJson() => {${[
    if(m.isInternal) "'id': id.hexString"
    else "'id': id",
    ...m.dataFields.map((f) => f.isNullable
        ? "if(${f.snapName} != null) '${f.remote}': ${f.jsonEncoder}"
        : "'${f.remote}': ${f.jsonEncoder}")
    ].join(',')}'
  '};';
}

extension ModelGeneratorX on Model {

  String get updateParams => dataFields.isEmpty ? '' :
    '{${dataFields.map((f) => f.updateParameters).expand((e) => e).join(',\n')}}';

  String get updateConstructor => [
    if(isNotSingleton) 'id',
    ...dataFields.map((f) => f.isNullable
        ? f.isModel
          ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})?.id'
          : '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
        : f.isModel
          ? '(${f.name} ?? this.${f.name}).id'
          : '(${f.name} ?? this.${f.name})')
  ].join(',');
}

extension ModelFieldX on ModelField {

  String get abstractGetter => '$snapType$nullable get $snapName;';

  List<String> get accessors => [
    getter,
    if(isModel)
      snapGetter,
    if(isNotHasMany)
      setter
  ];

  String get getter {
    if(isHasMany) {
      final type = '${model.name}${linkedModel?.namePlural}';
      return '$type get $name => $type._(_feature, this);';
    }
    final v = '_feature.get(this).$snapName';
    final m = typeModel;
    if(m != null) {
      if(m.isSingleton) {
        return isNullable
          ? '$refType? get $name => ($v != null) ? $refType._(_feature) : null;'
          : '$refType get $name => $refType._(_feature);';
      }
      return isNullable
        ? '$refType? get $name {'
          '  final id = $v;'
          '  return (id != null) ? $refType._(_feature, id) : null;'
          '}'
        : '$refType get $name => $refType._(_feature, $v);';
    }
    final override = isOverride ? '@override ' : '';
    return '$override$type$nullable get $name => $v;';
  }

  String get snapGetter {
    return '$snapType$nullable get $snapName => _feature.get(this).$snapName;';
  }

  String get setter => isNullable
      ? 'set $name($refType? value) => (value != null)'
        '  ? update($name: value)'
        '  : update(${name}AsNull: true);'
      : 'set $name($refType value) => update($name: value);';

  String get snapFieldDeclaration {
    if(isHasMany) {
      throw Exception('tried to call modelDeclaration on hasMany field ($this)');
    } else {
      final type = typeModel?.idType ?? this.type;
      return 'final $type$nullable $snapName;';
    }
  }

  List<String> get updateParameters => [
    '$keyType? $name',
    if(isNullable) 'bool ${name}AsNull=false'
  ];

  String get jsonType => typeModel?.idType ?? type;

  String get jsonDecoder => _jsonDecoder();
  String _jsonDecoder([String? v]) {
    v ??= "data['$remote']";
    return model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonDecoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? (isClass ? '$type.fromJson($v)' : null)
        ?? ((isModel && typeModel!.isInternal) ? '${typeModel!.keyName}.value($v)' : null)
        ?? (isModel ? v : null)
        ?? (isHasMany ? (throw Exception('jsonDecoder on isMany field')) : null)
        ?? (isDateTime ? 'DateTime.parse($v)' : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? fromJsonCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String get jsonEncoder {
    final v = snapName;
    final e = model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonEncoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.encoder?.replaceAll('\$', v)
        ?? (isHasMany ? (throw Exception('jsonEncoder on isMany field')) : null)
        ?? (jsonEncode?.replaceAll('\$', v))
        ?? (isClass ? '$v$nullable.toJson()' : null)
        ?? ((isModel && typeModel!.isInternal) ? '$v$nullable.hexString' : null)
        ?? (isModel ? v : null)
        ?? (isObjectId ? '$v$nullable.hexString' : null)
        ?? (isDateTime ? '$v$nullable.millisecondsSinceEpoch' : null)
        ?? (isPrimitive ? v : null)
        ?? (isInPart ? '$v$nullable.toJson()' : null)
        ?? (isEnum ? '$v$nullable.name' : null)
        ?? 'jsonValue($v)';
    return isNullable && isInitialized ? '$e ?? $initializer' : e;
  }

  String fromJsonCollection(String value) {
    if(isNotParameterized) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ModelField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final fromJson = field._jsonDecoder('v');
    if(isIterable) return '($value ?? []).map((v) => $fromJson).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $fromJson).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $fromJson).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $fromJson)).cast<$parameter>()';
    throw '$this is not a collection';
  }
}